import 'package:flutter/foundation.dart';
import 'package:result_monad/result_monad.dart';
import 'package:shared_preferences/shared_preferences.dart';

import '../friendica_client.dart';
import '../globals.dart';
import '../models/credentials.dart';
import '../models/exec_error.dart';
import 'secrets_service.dart';

class AuthService extends ChangeNotifier {
  FriendicaClient? _friendicaClient;
  bool _loggedIn = false;

  Result<FriendicaClient, ExecError> get currentClient {
    if (_friendicaClient == null) {
      return Result.error(ExecError(
        type: ErrorType.authentication,
        message: 'Not logged in',
      ));
    }

    return Result.ok(_friendicaClient!);
  }

  bool get loggedIn => _loggedIn && _friendicaClient != null;

  Future<bool> getStoredLoginState() async {
    final prefs = await SharedPreferences.getInstance();
    return prefs.getBool('logged-in') ?? false;
  }

  FutureResult<FriendicaClient, ExecError> signIn(
      Credentials credentials) async {
    final client = FriendicaClient(credentials: credentials);
    final result = await client.getMyProfile();
    if (result.isFailure) {
      return result.errorCast();
    }

    getIt<SecretsService>().storeCredentials(client.credentials);
    await _setLoginState(true);
    _friendicaClient = client;
    notifyListeners();
    return Result.ok(client);
  }

  Future signOut() async {
    print('Sign out');
    await _setLoginState(false);
    _friendicaClient = null;
    notifyListeners();
  }

  Future clearCredentials() async {
    _friendicaClient = null;
    await _setLoginState(false);
    notifyListeners();
  }

  Future<void> _setLoginState(bool state) async {
    final prefs = await SharedPreferences.getInstance();
    await prefs.setBool('logged-in', state);
    _loggedIn = state;
  }
}
